{-# LANGUAGE TypeSynonymInstances #-}

import Control.Monad (when)
import Foreign hiding (rotate)
import Foreign.C (CUChar)
import Foreign.C.String
import Foreign.ForeignPtr (withForeignPtr)
import Graphics.Rendering.OpenGL.Raw
import Graphics.UI.GLUT hiding (rotate, translate, compileShader, RGBA)
import Data.Array.Repa.IO.DevIL
import Data.Array.Repa.Repr.ForeignPtr

type R = Double
data V = V !R !R !R deriving Show
data M = M !V !V !V deriving Show

class ApproxEq t where (=~) :: t -> t -> Bool
instance ApproxEq R where x =~ y = abs (x - y) < 1e-6
instance ApproxEq V where (V u v w) =~ (V x y z) = and (zipWith (=~) [u,v,w] [x,y,z])
instance ApproxEq M where (M u v w) =~ (M x y z) = and (zipWith (=~) [u,v,w] [x,y,z])

dot :: V -> V -> R
dot (V u v w) (V x y z) = u * x + v * y + w * z

opposite :: V -> V
opposite (V u v w) = (V (-u) (-v) w)

identity :: M
identity = M (V 1 0 0) (V 0 1 0) (V 0 0 1)

transpose :: M -> M
transpose (M (V a b c) (V d e f) (V g h i)) = M (V a d g) (V b e h) (V c f i)

multiply :: M -> M -> M
multiply m n = multiply' m (transpose n)
  where
    multiply' (M a b c) (M p q r) = M (V (dot a p) (dot a q) (dot a r)) (V (dot b p) (dot b q) (dot b r)) (V (dot c p) (dot c q) (dot c r))

mul :: M -> V -> V
mul (M a b c) v = V (dot a v) (dot b v) (dot c v)

translate :: V -> M
translate (V x y w) = M (V 1 0 (x/w)) (V 0 1 (y/w)) (V 0 0 1)

rotate :: R -> V -> M
rotate a o = translate o `multiply` rotate' (2 * pi * a) `multiply` translate (opposite o)
  where
    rotate' a' = M (V c s 0) (V (-s) c 0) (V 0 0 1)
      where
        c = cos a'
        s = sin a'

distance :: M -> R
distance m = let V u v w = mul m (V 0 0 1) in sqrt ((u/w)*(u/w) + (v/w)*(v/w))

m2, m3, m6 :: V -> M
m2 = rotate (1/2)
m3 = rotate (1/3)
m6 = rotate (1/6)

type C = V
data P = P C C C C deriving Show

instance ApproxEq P where P a b c d =~ P x y z w = a =~ x && b =~ y && c =~ z && d =~ w

p2, p3, p6 :: P -> P
p2 (P a b c d) = P b a d c
p3 (P a b c d) = P b c a d
p6 (P a b c d) = P d b a c

data Butterfly = B V V V M P deriving Show

red, yellow, green, blue :: C
red    = V 0.843 0.129 0.000
yellow = V 1.000 0.725 0.153
green  = V 0.500 0.643 0.277
blue   = V 0.373 0.514 0.722

near :: Butterfly -> Bool
near (B _ _ _ m _) = distance m < 4

butterfly :: Butterfly
butterfly = B (V 0 0 1) (V (3/4) 0 1) (V (3/4) (sqrt 3/4) 1) identity (P red yellow green blue)

instance ApproxEq Butterfly where
  B u v w m p =~ B x y z n q = p =~ q && u =~ x && v =~ y && w =~ z && m =~ n

neighbours :: Butterfly -> [Butterfly]
neighbours (B u v w n p) =
  [ let m = m6 u in B (mul m u) (mul m v) (mul m w) (multiply m n) (p6 p)
  , let m = m2 v in B (mul m u) (mul m v) (mul m w) (multiply m n) (p2 p)
  , let m = m3 w in B (mul m u) (mul m v) (mul m w) (multiply m n) (p3 p)
  ]

swarm :: [Butterfly]
swarm = closure near (=~) neighbours butterfly

closure :: (t -> Bool) -> (t -> t -> Bool) -> (t -> [t]) -> t -> [t]
closure p e f x = closure' [] [x]
  where
    closure' _ [] = []
    closure' old xs = xs ++ closure' (old ++ new) new
      where
        new = [ y | y <- concatMap f xs, p y, not (elemBy e y old) ]

elemBy :: (t -> t -> Bool) -> t -> [t] -> Bool
elemBy eq x [] = False
elemBy eq x (y:ys)
  | eq x y = True
  | otherwise = elemBy eq x ys

class GLPoke t where glPoke :: t -> Ptr t -> IO (Ptr t)
instance GLPoke t => GLPoke [t] where
  glPoke [] p = return p
  glPoke (x:xs) p = glPoke x (castPtr p) >>= glPoke xs . castPtr

instance GLPoke GLfloat where glPoke x p = poke p x >> return (advancePtr p 1)
instance GLPoke R where
  glPoke x p = do
    p <- glPoke (realToFrac x :: GLfloat) (castPtr p)
    return (castPtr p)
instance GLPoke V where
  glPoke (V x y z) p = do
    p <- glPoke x (castPtr p)
    p <- glPoke y p
    p <- glPoke z p
    return (castPtr p)
instance GLPoke M where
  glPoke (M x y z) p = do
    p <- glPoke x (castPtr p)
    p <- glPoke y p
    p <- glPoke z p
    return (castPtr p)
instance GLPoke P where
  glPoke (P x y z _) p = do
    p <- glPoke x (castPtr p)
    p <- glPoke y p
    p <- glPoke z p
    return (castPtr p)
instance GLPoke Butterfly where
  glPoke (B _ _ _ m q) p = do
    p <- glPoke m (castPtr p)
    p <- glPoke q (castPtr p)
    return (castPtr p)

vert = unlines
  [ "#version 400 core"
  , "layout(location = 0) in vec3 t0a;"
  , "layout(location = 1) in vec3 t0b;"
  , "layout(location = 2) in vec3 t0c;"
  , "layout(location = 3) in vec3 c0a;"
  , "layout(location = 4) in vec3 c0b;"
  , "layout(location = 5) in vec3 c0c;"
  , "out mat3 t1;"
  , "out mat3 c1;"
  , "void main() {"
  , "  t1 = transpose(mat3(t0a, t0b, t0c));"
  , "  c1 = mat3(c0a, c0b, c0c);"
  , "}"
  ]

geom = unlines
  [ "#version 400 core"
  , "layout(points) in;"
  , "layout(triangle_strip, max_vertices = 4) out;"
  , "uniform mat4 mvp;"
  , "in mat3 t1[1];"
  , "in mat3 c1[1];"
  , "smooth out vec2 t;"
  , "flat out mat3 c;"
  , "void main() {"
  , "  c = c1[0];"
  , "  vec3 p;"
  , "  vec2 tt;"
  , "  tt = vec2(0.0, 0.0);"
  , "  t = tt;"
  , "  p = t1[0] * vec3(tt * vec2(1.0, 0.4330127018922193), 1.0);"
  , "  gl_Position = vec4(mvp * vec4(p.x/p.z, p.y/p.z, 0.0, 1.0));"
  , "  EmitVertex();"
  , "  tt = vec2(1.0, 0.0);"
  , "  t = tt;"
  , "  p = t1[0] * vec3(tt * vec2(1.0, 0.4330127018922193), 1.0);"
  , "  gl_Position = vec4(mvp * vec4(p.x/p.z, p.y/p.z, 0.0, 1.0));"
  , "  EmitVertex();"
  , "  tt = vec2(0.0, 1.0);"
  , "  t = tt;"
  , "  p = t1[0] * vec3(tt * vec2(1.0, 0.4330127018922193), 1.0);"
  , "  gl_Position = vec4(mvp * vec4(p.x/p.z, p.y/p.z, 0.0, 1.0));"
  , "  EmitVertex();"
  , "  tt = vec2(1.0, 1.0);"
  , "  t = tt;"
  , "  p = t1[0] * vec3(tt * vec2(1.0, 0.4330127018922193), 1.0);"
  , "  gl_Position = vec4(mvp * vec4(p.x/p.z, p.y/p.z, 0.0, 1.0));"
  , "  EmitVertex();"
  , "  EndPrimitive();"
  , "}"
  ]

frag = unlines
  [ "#version 400 core"
  , "uniform sampler2D tex;"
  , "flat in mat3 c;"
  , "smooth in vec2 t;"
  , "out layout(location = 0, index = 0) vec4 f;"
  , "void main() {"
  , "  vec4 w = texture(tex, t * vec2(1.0, -1.0) + vec2(0.0, 1.0));"
  , "  if (w.a < 0.5) {"
  , "    discard;"
  , "  } else if (w.r > 0.5 && w.g > 0.5 && w.b > 0.5) {"
  , "    f = vec4(1.0, 1.0, 1.0, 1.0);"
  , "  } else if (w.b > 0.5) {"
  , "    f = vec4(c[0], 1.0);"
  , "  } else if (w.r > 0.5) {"
  , "    f = vec4(c[1], 1.0);"
  , "  } else if (w.g > 0.5) {"
  , "    f = vec4(c[2], 1.0);"
  , "  } else {"
  , "    f = vec4(0.0, 0.0, 0.0, 1.0);"
  , "  }"
  , "}"
  ]

main = do
  _ <- getArgsAndInitialize
  _ <- createWindow "butterfly"
  loadTexture
  let count = length swarm
      stride = (3 * 3 + 3 * 3) * 4
      bytes = count * fromIntegral stride
  allocaBytes bytes $ \p -> do
    glPoke swarm p
    vbo <- with 0 $ \q -> glGenBuffers 1 q >> peek q
    glBindBuffer gl_ARRAY_BUFFER vbo
    glBufferData gl_ARRAY_BUFFER (fromIntegral bytes) p gl_STATIC_DRAW
    vao <- with 0 $ \q -> glGenVertexArrays 1 q >> peek q
    glBindVertexArray vao
    glEnableClientState gl_VERTEX_ARRAY
    glVertexPointer 2 gl_FLOAT 0 nullPtr
    glVertexAttribPointer 0 3 gl_FLOAT (fromIntegral gl_FALSE) stride (plusPtr nullPtr (0 * 3 * 4))
    glVertexAttribPointer 1 3 gl_FLOAT (fromIntegral gl_FALSE) stride (plusPtr nullPtr (1 * 3 * 4))
    glVertexAttribPointer 2 3 gl_FLOAT (fromIntegral gl_FALSE) stride (plusPtr nullPtr (2 * 3 * 4))
    glVertexAttribPointer 3 3 gl_FLOAT (fromIntegral gl_FALSE) stride (plusPtr nullPtr (3 * 3 * 4))
    glVertexAttribPointer 4 3 gl_FLOAT (fromIntegral gl_FALSE) stride (plusPtr nullPtr (4 * 3 * 4))
    glVertexAttribPointer 5 3 gl_FLOAT (fromIntegral gl_FALSE) stride (plusPtr nullPtr (5 * 3 * 4))
    glEnableVertexAttribArray 0
    glEnableVertexAttribArray 1
    glEnableVertexAttribArray 2
    glEnableVertexAttribArray 3
    glEnableVertexAttribArray 4
    glEnableVertexAttribArray 5
  program <- compileProgram
  glUseProgram program
  let s = 2
      l = -s
      r = s
      t = -s
      b = s
      n = 1
      f = 2
      ortho = [ 2 / (r - l), 0, 0, (r + l) / (r - l)
              , 0, 2 / (t - b), 0, (t + b) / (t - b)
              , 0, 0, -2 / (f - n), -(f + n) / (f - n)
              , 0, 0, 0, 1 ]
  withArray ortho $ \p -> do
    loc <- withCString "mvp" $ glGetUniformLocation program
    glUniformMatrix4fv loc 1 (fromIntegral gl_FALSE) p
  loc <- withCString "tex" $ glGetUniformLocation program
  glUniform1i loc 0
  glClearColor 0.5 0.5 0.5 1
  displayCallback $= do
    glClear gl_COLOR_BUFFER_BIT
    glDrawArrays gl_POINTS 0 (fromIntegral count)
    swapBuffers
    reportErrors
  mainLoop

loadTexture = do
  RGBA img <- runIL $ readImage "butterfly.png"
  withForeignPtr (toForeignPtr img) $ \p -> do
    tex <- with 0 $ \q -> glGenTextures 1 q >> peek q
    glBindTexture gl_TEXTURE_2D tex
    glTexImage2D gl_TEXTURE_2D 0 (fromIntegral gl_RGBA) 1024 1024 0 gl_RGBA gl_UNSIGNED_BYTE p
    glGenerateMipmap gl_TEXTURE_2D
    
compileProgram = do
  program <- glCreateProgram
  compileShader program gl_VERTEX_SHADER vert
  compileShader program gl_GEOMETRY_SHADER geom
  compileShader program gl_FRAGMENT_SHADER frag
  glLinkProgram program
  debugProgram program
  return program

compileShader program t src = do
  shader <- glCreateShader t
  withCString src $ \srcp -> with srcp $ \srcpp -> glShaderSource shader 1 srcpp nullPtr
  glCompileShader shader
  glAttachShader program shader
  glDeleteShader shader

debugProgram program = do
  if (program /= 0)
    then do
      linked <- with 0 $ \p -> glGetProgramiv program gl_LINK_STATUS p >> peek p
      when (linked /= fromIntegral gl_TRUE) $ putStrLn "link failed"
      len <- with 0 $ \p -> glGetProgramiv program gl_INFO_LOG_LENGTH p >> peek p
      allocaBytes (fromIntegral len + 1) $ \p -> glGetProgramInfoLog program len nullPtr p >> pokeByteOff p (fromIntegral len) (0 :: CUChar) >> peekCString p >>= putStrLn
  else putStrLn "no program"
